import datetime
import sys
from sys import platform
import socket
import re
import configparser
import os
import subprocess
import pefile
from py_logging.cestel_logging import log, log_exception
import logging
import requests

if platform == "win32":
  import winreg

from datetime import timedelta
from lxml import etree
from generic_module import Module
 
class StatusModule(Module):
  def __init__(self, name, host, update_channel, time_interval, version, url, protocol, root_path, ntp_conf_path, ovpn_path, save_curr_to_disk, vpn_host):
    Module.__init__(self, name, root_path)
    self.time_interval = time_interval
    self.host = host
    self.update_channel = update_channel
    self.version = version
    self.url = url
    self.protocol = protocol 
    self.ntp_conf_path = ntp_conf_path
    self.ovpn_path = ovpn_path
    self.save_curr_to_disk = save_curr_to_disk
    self.vpn_host = vpn_host
  
  def swmCurSite(self, loc):
    try:
      cfg = configparser.ConfigParser(strict=False)  # Ignore potential duplicate keys in conf
      cfg.read(loc)
      return cfg.get('global', 'site')
    except:
      log(self.logger, 'Failed to obtain site name from {}'.format(loc), sev=logging.WARNING)
      return 'unknown'
  
  def VPNstatus(self):  
    if platform == "win32":
      ping = subprocess.Popen(["ping", "-n", "1", "-w", "2", self.vpn_host],stdout = subprocess.PIPE).communicate()[0]
    else:
      ping = subprocess.Popen(["ping", "-c", "1", "-w", "2", self.vpn_host],stdout = subprocess.PIPE).communicate()[0]
    if ('unreachable' not in str(ping)) and ('timed' not in str(ping)) and ('failure' not in str(ping)): 
      return "Connected"
    return "Not connected"
    
  def w32ovpn_version(self):  
    sp = subprocess.Popen([self.ovpn_path, "--version"], stdout=subprocess.PIPE)
    output = sp.communicate()[0]
    return output.split(b" ")[1].decode()  # Output of the application is binary and thus has to be split by "binary spaces". We could also decode before splitting, but that's probably redundant.

  def getUniXhostname(self):
    sp = subprocess.Popen(["hostname"], stdout=subprocess.PIPE)
    output = sp.communicate()[0]
    return output.strip()

  def get_file_info(self, filename):
    # Version is packed into 2 integers, with each containing 2 values. Thus these convenience functions return least/most significant words.
    def lsw(dword):
      return dword & 0x0000ffff
    def msw(dword):
      return dword >> 16

    pe = pefile.PE(filename)
    ms = pe.VS_FIXEDFILEINFO[0].ProductVersionMS
    ls = pe.VS_FIXEDFILEINFO[0].ProductVersionLS
    return '{0}.{1}.{2}.'.format(msw(ms), lsw(ms), msw(ls))

  def get_registry_value(self, key, subkey, value):
    key = getattr(_winreg, key)
    handle = winreg.OpenKey(key, subkey)
    (value, type) = winreg.QueryValueEx(handle, value)
    return value

  def getRouterInfo(self):
    # Define "constants" for convenience
    ROUTER, IMEI, FREQ_BAND, IMSI, LAT, LON, RSRQ = 'router', 'gsm_imei', 'gsm_frequency_band', 'gsm_imsi', 'gps_lat', 'gps_lon', 'gsm_signal_quality'
    ports = ('80', '9191')  # Ports that we want to check
    router_id = {'Sierra': 'Sierra', 'Teltonika': '/cgi-bin/luci'}  # Teltonika has a redirect, which holds the actual data
    data = {IMSI: 'No IMSI'}
    for port in ports:
      ip = 'http://192.168.4.100:{}'.format(port)  # IP as a variable so that code can be tested anywhere without modifications.
      try:
        req = requests.get(ip, timeout=10)
      except requests.exceptions.ConnectionError:  # If connection can't be established, continue with next port
        continue
      if not req.ok:
        log(self.logger, 'Router not found at {}: Error code {}. Trying next IP in {}'.format(ip, req.status_code, ', '.join(ports)))
        continue

      router = ''
      for type in router_id.values():
        if req.text.find(type) != -1:  # If the router's name exists in the entry page, we assume it's the router of that company
          router = type
          break

      if router == router_id['Sierra']:
        try:
          data[ROUTER] = '{}: {}'.format(router, re.findall(r'ALEOS Version (?:\d+\.?)+', req.text)[0])  # Exception should happen here if it does.
          # Obtain the data we're interested in
          with requests.Session() as s:
            login_data = '<request xmlns="urn:acemanager"><connect><login>user</login><password><![CDATA[12345]]></password></connect></request>'
            req = s.post('{}/xml/Connect.xml'.format(ip), data=login_data, timeout=10)
            if not req.ok:
              log(self.logger, 'Failed to log in to Sierra: {}'.format(req), sev=logging.WARNING)
              return
            # 7=Device model, 10=IMEI, 671=active frequency band, 785=IMSI, 902=latitude, 903=longitude, 10209=signal quality
            req = s.post('{}/cgi-bin/Embedded_Ace_Get_Task.cgi'.format(ip), data='7,10,671,785,902,903,10209', timeout=10)
            if not req.ok:
              log(self.logger, 'Failed to obtain information from Sierra: {}'.format(req), sev=logging.WARNING)
              return
          for val, key in zip(req.text.split('!')[:-1], (ROUTER, IMEI, FREQ_BAND, IMSI, LAT, LON, RSRQ)):  # Since it ends with a !, last element is empty
            data[key] = val.split('=')[1]
          try:
            log(self.logger, 'lat: {}, lon: {}'.format(data[LAT], data[LAT]))
            data[LAT] = '{}.{}'.format(data[LAT][:-5], data[LAT][-5:])
            data[LON] = '{}.{}'.format(data[LON][:-5], data[LON][-5:])
          except:
            log(self.logger, 'Could not convert latitude ({}) and/or longitude ({}) to decimal.'.format(data.get(LAT), data.get(LON)), sev=logging.WARNING)
        except IndexError:  # The old version of Sierra does not use ALEOS and thus regex match fails
          log(self.logger, 'Failed to obtain router information for a Sierra router. Falling back to "old" way.')
          data[ROUTER] = '{} {}'.format(router, data[ROUTER] if ROUTER in data else '')  # Add the router name before the type.
          data[LAT], data[LON] = self.gpsLoc()
          data[IMSI] = self.SierraImsi('12345')

      elif router == router_id['Teltonika']:
        req = requests.get('{}{}'.format(ip, router_id["Teltonika"]), timeout=10)  # This request will return status code 403, but still enough information to identify the router.
        if req.text.find('Teltonika') != -1:  # We double check if it really is Teltonika because their entry page has no unique identification available; we assume it is from redirect path
          router = 'Teltonika'
          payload = {
            'username': 'admin',
            'password': 'Siwim323'
          }

          data[ROUTER] = router
          data_dict = requests.post('{}/cgi-bin/luci/admin/status/netinfo/mobile?status=1'.format(ip), data=payload, timeout=10).json()
          data[IMEI] = data_dict['gimei']
          data[FREQ_BAND] = data_dict['bands']
          data[IMSI] = data_dict['gimsi']
          data[RSRQ] = data_dict['grsrq']
          data_dict = requests.post('{}/cgi-bin/luci/admin/services/gps/get_cord'.format(ip), data=payload, timeout=10).json()
          data[LAT] = data_dict['latitude']
          data[LON] = data_dict['longitude']
        else:
          log(self.logger, 'Unrecognised router (assumed Teltonika), IMSI and GPS information not obtained.', sev=logging.WARNING)
      else:
        log(self.logger, 'Unrecognised router (could not find a match), IMSI and GPS information not obtained.', sev=logging.WARNING)
      break  # We've gotten our router information, so break the loop.
    log(self.logger, 'Router data: {}'.format(data), sev=logging.DEBUG)
    return data

  def gpsLoc(self):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(5.0)
    data = 'No GPS data'
    try:
      s.connect(('192.168.4.100', 9494))
      data = s.recv(1024).decode()
      s.close()
    except:
      log(self.logger, 'Obtaining legacy SierraGpsLoc failed', sev=logging.DEBUG, exc_info=True)

    p = re.compile('GPGGA,([.\d]*),([.\d]*),([NS]),([.\d]*),([EW])')
    a = p.search(data)

    if a == None:
      return (0, 0)
    else:
      latd = float(a.group(2)[:2])
      lond = float(a.group(4)[:3])
      latm = float(a.group(2)[2:])
      lonm = float(a.group(4)[3:])
      
      lat = latd + (latm / 60)
      lon = lond + (lonm / 60)

      if a.group(3) != 'N':
        lat = -lat
      if a.group(5) != 'E':
        lon = -lon
      return (lat, lon)

  def _recvfind(self, sock, n, to_sec=5):
    d = ''
    absto = datetime.datetime.now() + timedelta(seconds=to_sec)
    while (-1 == d.find(n)) and (datetime.datetime.now() < absto):
      d += sock.recv(256).decode(errors='replace')
    return re.sub(r'[^a-zA-Z0-9 \n]', '', d)

  def SierraImsi(self, pwd):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(5.0)
    imsi = 'No IMSI'
    try:
      s.connect(('192.168.4.100', 2332))
      data = self._recvfind(s, ':')
      if data.find('login') >= 0:
        #print 'GX'
        s.sendall('user\r\n'.encode())
        data = self._recvfind(s, ':')
        #print 'MP'
      s.sendall((pwd + '\r\n').encode())
      data = self._recvfind(s, 'OK')
      s.sendall('at+cimi?\r\n'.encode())
      data = self._recvfind(s, 'OK')
      data = data.split()
      if len(data) > 1: imsi = data[-2]
      s.close()
    except:
      log(self.logger, 'Obtaining legacy SierraImsi failed', sev=logging.DEBUG, exc_info=True)
    return imsi

  def swmiver(self):
    return ".".join(str(x) for x in self.version)

  def ntpsvr(self, loc):
    fn = loc
    if not os.path.exists(fn):
      return 'missing'
    f = open(fn)
    try:
      lines = f.readlines()
    finally:
      f.close()
    for l in lines:
      if l.strip().startswith('server'):
        return l.strip().split()[1] 
    for l in lines:
      if l.strip().startswith('pool'):
        return l.strip().split()[1] 
    return 'unknown'

  def swmstat(self):
    kvp = {}
    try:
      kvp['python_version']      = str(sys.version_info[0])+'.'+str(sys.version_info[1])+'.'+str(sys.version_info[2])
    except:
      log(self.logger, 'Python version not saved.', sev=logging.WARNING)
    try:
      kvp['siwim_version'] = '4' if sys.getwindowsversion()[0] == 10 else '3'
      if sys.getwindowsversion()[0] == 10 and sys.version_info[0] == 2:  # Log error if the system is MK4, but is running on Python2
        log(self.logger, 'Version mismatch! Windows 10 should use Python3.', sev=logging.WARNING)
    except:
      log(self.logger, 'SiWIM version unavailable.', sev=logging.WARNING)
    try:
      kvp['update_channel'] = self.update_channel
    except:
      log(self.logger, 'Update channel not saved.')
    try:
      kvp['usb_info'] = subprocess.check_output([sys.executable, 'get_usb_info.py'], stderr=open(os.devnull, 'w')).decode()
    except:
      log(self.logger, 'USB info not saved.', sev=logging.WARNING)
    try:
      kvp['timestamp']  = datetime.datetime.now().strftime('%Y-%m-%d %H:%M')  
    except:
      log(self.logger, 'Timestamp not saved.', sev=logging.WARNING)
    try:
      if platform == "win32":
        kvp['computer_name']   = os.getenv('COMPUTERNAME')
      else:
        kvp['computer_name']   = self.getUniXhostname()
    except:
      log(self.logger, 'Computer name not saved.', sev=logging.WARNING)
    # To reduce repeated code, map values of partitions depending on platform and save them to partitions dict.
    if platform == 'win32':
      partitions = {'disk_c': 'c:', 'disk_d': 'd:', 'disk_e': 'e:'}
    else:
      partitions = {'disk_c': '/', 'disk_d': self.root_path, 'disk_e': '/'}
    for key, p in partitions.items():
      try:
        kvp['{0}_free'.format(key)] = self.get_free_mb(p)
      except:
        log(self.logger, 'Device {0} not found.'.format(p), sev=logging.WARNING)
    try:
      if platform == "win32":
        kvp['siwim_site_name']       = self.swmCurSite(self.root_path + "/siwim_e/conf/siwim.conf")
      else:
        kvp['siwim_site_name']       = self.swmCurSite(self.root_path + "/siwim_e/conf/siwim.conf")
    except:
      log(self.logger, 'Site name not saved.', sev=logging.WARNING)
    try:
      kvp['vpn_status']       = self.VPNstatus()
    except:
      log(self.logger, 'VPN status not saved.', sev=logging.WARNING)
    try:
      kvp['siwim_e_version'] = self.get_file_info(self.root_path + '/siwim_e/bin/siwim_mcp.exe')
    except:
      log(self.logger, 'E version not saved.', sev=logging.WARNING)
    try:
      kvp['siwim_i_version']    = self.swmiver()
    except:
      log(self.logger, 'I version not saved.', sev=logging.WARNING)
    for key, p in partitions.items():
      try:
        kvp['{0}_total'.format(key)] = self.get_total_mb(p)
      except:
        log(self.logger, 'Total MB not saved for {0}.'.format(p), sev=logging.WARNING)
    try:
      tmp = self.getRouterInfo()
      for key, val in tmp.items():
        kvp[key] = val
    except:
      log(self.logger, 'GSM IMSI and location not saved.', sev=logging.WARNING)
    try:
      if self.ntp_conf_path == "" or self.ntp_conf_path is None:
        if platform == "win32":
          kvp['ntp_server']        = self.ntpsvr("c:/windows/system32/drivers/etc/ntp.conf") 
        else:
          kvp['ntp_server']        = self.ntpsvr("/etc/ntp.conf") 
      else:
        kvp['ntp_server']        = self.ntpsvr(self.ntp_conf_path) 
    except:
      log(self.logger, 'NTP server not saved', sev=logging.WARNING)
    try:
      #if platform == "win32":
      kvp['openvpn_version']     = self.w32ovpn_version()
      #else:
        #kvp['openvpn_version']     = self.uniX32ovpn_version()
    except:
      log(self.logger, 'OpenVPN not saved.', logging.WARNING)
    
    
    for key, val in self.upstream_info.items():
      try:
        kvp[key]     = val
      except:
        log(self.logger, str(key) + ' not saved.', logging.WARNING)
    return kvp
  
  def run(self):
    self.alive = True
    self.end = False
    try:
      while True:
        if self.end:
          self.alive = False
          log(self.logger, 'Thread closed correctly.', sev=logging.DEBUG)
          return
        try:
          kvp = self.swmstat()
          root = etree.Element("swmstat") 
          
          voltage_present = False
          for key, val in kvp.items():
            if isinstance(val, dict):
              nest = root.find(key[1] + "s")
              if nest == None:
                nest = etree.Element(key[1] + "s")
                root.append(nest)
              node = etree.Element(key[1])
              nest.append(node)
              for ky in val.keys():
                node.attrib[ky] = val[ky]
            else:
              #ugly hack, but alas
              nn = ""
              if isinstance(key, tuple):
                nn = key[1]
              else:
                nn = key
              node = etree.Element(nn) 
              node.text = str(val)
              root.append(node)
              if nn == "siwim_system_voltage":
                voltage_present = True
          #for non-standard systems that don't yield voltage info, despite it being mandatory for a valid register event(alive)
          #if this piece of info ever stops being mandatory (i see no reason why it is), remove this hack
          if not voltage_present:
            node = etree.Element("siwim_system_voltage")
            node.text = "-1"
            root.append(node)
          if self.save_curr_to_disk == 1:
            log(self.logger, 'Writing status to disk ...')
            fd = open("log/swmstat.xml", "w")
            fd.write(etree.tostring(root).decode())
            fd.close()
          resp = requests.post('{0}://{1}{2}'.format(self.protocol, self.host, self.url), {'sys': kvp["computer_name"].lower(), 'msg': etree.tostring(root).decode()}, timeout=10, verify=False)
          if resp.status_code != 204:  # Log, if response isn't empty.
            log(self.logger, 'Response: {0}'.format(resp.text), sev=logging.WARNING)
        except socket.error as e:
          log(self.logger, 'Connection attempt failed ({0}).'.format(e), sev=logging.WARNING)
        except:
          log_exception(self.logger, 'General error:')
        #time.sleep(self.time_interval)
        self.zzzzz(self.time_interval)
    except:
      log_exception(self.logger, 'Fatal error:')
      self.alive = False
